﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Threading;

namespace WawaSoft.Net
{
    public class WaSocket : IDisposable
    {
        #region fields
        private static object s_InternalSyncObject;
        internal static bool s_Initialized;

        private bool willBlock;
        private bool willBlockInternal;
        private bool isListening;
        private AddressFamily addressFamily;
        private SocketType socketType;
        private ProtocolType protocolType;

        private SafeCloseSocket m_Handle;
        internal EndPoint m_RightEndPoint;
        internal EndPoint m_RemoteEndPoint;
        private bool m_IsDisconnected;
        private int m_CloseTimeout;
        private bool m_IsConnected;
        private int m_IntCleanedUp;
        #endregion

        #region static methods
        private static void InitializeSockets()
        {
            if (!s_Initialized)
            {
                lock (InternalSyncObject)
                {
                    if (!s_Initialized)
                    {
                        WSAData lpWSAData = new WSAData();
                        if (OSSOCK.WSAStartup(0x202, out lpWSAData) != SocketError.Success)
                        {
                            throw new SocketException();
                        }
                        s_Initialized = true;
                    }
                }
            }
        }

        private static object InternalSyncObject
        {
            get
            {
                if (s_InternalSyncObject == null)
                {
                    object obj2 = new object();
                    Interlocked.CompareExchange(ref s_InternalSyncObject, obj2, null);
                }
                return s_InternalSyncObject;
            }
        } 
        #endregion

        #region public contructs
        public WaSocket(AddressFamily addressFamily, SocketType socketType, ProtocolType protocolType)
        {
            this.willBlock = true;
            this.willBlockInternal = true;
            this.m_CloseTimeout = -1;

            InitializeSockets();
            this.m_Handle = SafeCloseSocket.CreateWSASocket(addressFamily, socketType, protocolType);
            if (this.m_Handle.IsInvalid)
            {
                throw new SocketException();
            }
            this.addressFamily = addressFamily;
            this.socketType = socketType;
            this.protocolType = protocolType;
        } 
        #endregion

        #region private contructs
        private WaSocket(SafeCloseSocket fd)
        {
            this.willBlock = true;
            this.willBlockInternal = true;
            this.m_CloseTimeout = -1;

            InitializeSockets();
            if ((fd == null) || fd.IsInvalid)
            {
                throw new ArgumentException("net_InvalidSocketHandle");
            }
            this.m_Handle = fd;
            this.addressFamily = AddressFamily.Unknown;
            this.socketType = SocketType.Unknown;
            this.protocolType = ProtocolType.Unknown;
        } 
        #endregion

        #region Sync Methods
        #region sync Bind
        public void Bind(EndPoint localEP)
        {
            if (this.CleanedUp)
            {
                throw new ObjectDisposedException(base.GetType().FullName);
            }
            if (localEP == null)
            {
                throw new ArgumentNullException("localEP");
            }
            EndPoint endPointSnapshot = localEP;
            SocketAddress socketAddress = endPointSnapshot.Serialize();
            this.DoBind(endPointSnapshot, socketAddress);
        }

        private void DoBind(EndPoint endPointSnapshot, SocketAddress socketAddress)
        {
            byte[] m_buffer;
            int m_Size;
            m_buffer = GetM_buffer(socketAddress, out m_Size);

            if (OSSOCK.bind(this.m_Handle, m_buffer, m_Size) != SocketError.Success)
            {
                SocketException socketException = new SocketException();
                this.UpdateStatusAfterSocketError(socketException);
                throw socketException;
            }
            if (this.m_RightEndPoint == null)
            {
                this.m_RightEndPoint = endPointSnapshot;
            }
        }
        #endregion

        #region Sync Listen
        public void Listen(int backlog)
        {
            if (this.CleanedUp)
            {
                throw new ObjectDisposedException(base.GetType().FullName);
            }
            if (OSSOCK.listen(this.m_Handle, backlog) != SocketError.Success)
            {
                SocketException socketException = new SocketException();
                this.UpdateStatusAfterSocketError(socketException);
                throw socketException;
            }
            this.isListening = true;
        }
        #endregion

        #region Sync Accept
        public WaSocket Accept()
        {
            if (this.CleanedUp)
            {
                throw new ObjectDisposedException(base.GetType().FullName);
            }
            if (this.m_RightEndPoint == null)
            {
                throw new InvalidOperationException("net_sockets_mustbind");
            }
            if (!this.isListening)
            {
                throw new InvalidOperationException("net_sockets_mustlisten");
            }
            if (this.m_IsDisconnected)
            {
                throw new InvalidOperationException("net_sockets_disconnectedAccept");
            }
            this.ValidateBlockingMode();
            SocketAddress socketAddress = this.m_RightEndPoint.Serialize();
            byte[] m_buffer;
            int m_Size;
            m_buffer = GetM_buffer(socketAddress, out m_Size);
            SafeCloseSocket fd = SafeCloseSocket.Accept(this.m_Handle, m_buffer, ref m_Size);
            if (fd.IsInvalid)
            {
                SocketException socketException = new SocketException();
                this.UpdateStatusAfterSocketError(socketException);
                throw socketException;
            }
            WaSocket retObject = this.CreateAcceptSocket(fd, this.m_RightEndPoint.Create(socketAddress), false);
            return retObject;
        }

        private WaSocket CreateAcceptSocket(SafeCloseSocket fd, EndPoint remoteEP, bool needCancelSelect)
        {
            WaSocket socket = new WaSocket(fd);
            return this.UpdateAcceptSocket(socket, remoteEP, needCancelSelect);
        }

        internal WaSocket UpdateAcceptSocket(WaSocket socket, EndPoint remoteEP, bool needCancelSelect)
        {
            socket.addressFamily = this.addressFamily;
            socket.socketType = this.socketType;
            socket.protocolType = this.protocolType;
            socket.m_RightEndPoint = this.m_RightEndPoint;
            socket.m_RemoteEndPoint = remoteEP;
            socket.SetToConnected();
            if (needCancelSelect)
            {
                //socket.UnsetAsyncEventSelect();
            }
            socket.willBlock = this.willBlock;
            if (this.willBlock != this.willBlockInternal)
            {
                socket.InternalSetBlocking(this.willBlock);
            }
            return socket;
        }
        #endregion

        #region sync receive
        public int Receive(byte[] buffer)
        {
            return this.Receive(buffer, 0, (buffer != null) ? buffer.Length : 0, SocketFlags.None);
        }
        public int Receive(byte[] buffer, int offset, int size, SocketFlags socketFlags)
        {
            SocketError error;
            int num = this.Receive(buffer, offset, size, socketFlags, out error);
            if (error != SocketError.Success)
            {
                throw new SocketException((int)error);
            }
            return num;
        }
        public unsafe int Receive(byte[] buffer, int offset, int size, SocketFlags socketFlags, out SocketError errorCode)
        {
            int num;
            if (this.CleanedUp)
            {
                throw new ObjectDisposedException(base.GetType().FullName);
            }
            if (buffer == null)
            {
                throw new ArgumentNullException("buffer");
            }
            if ((offset < 0) || (offset > buffer.Length))
            {
                throw new ArgumentOutOfRangeException("offset");
            }
            if ((size < 0) || (size > (buffer.Length - offset)))
            {
                throw new ArgumentOutOfRangeException("size");
            }
            this.ValidateBlockingMode();
            errorCode = SocketError.Success;
            if (buffer.Length == 0)
            {
                num = OSSOCK.recv(this.m_Handle.DangerousGetHandle(), null, 0, socketFlags);
            }
            else
            {
                fixed (byte* numRef = buffer)
                {
                    num = OSSOCK.recv(this.m_Handle.DangerousGetHandle(), numRef + offset, size, socketFlags);
                }
            }
            if (num == -1)
            {
                errorCode = (SocketError)Marshal.GetLastWin32Error();
                this.UpdateStatusAfterSocketError(errorCode);
                return 0;
            }
            return num;
        }
        #endregion

        #region sync send
        public int Send(byte[] buffer)
        {
            return this.Send(buffer, 0, (buffer != null) ? buffer.Length : 0, SocketFlags.None);
        }
        public int Send(byte[] buffer, int offset, int size, SocketFlags socketFlags)
        {
            SocketError error;
            int num = this.Send(buffer, offset, size, socketFlags, out error);
            if (error != SocketError.Success)
            {
                throw new SocketException((int)error);
            }
            return num;
        }
        public unsafe int Send(byte[] buffer, int offset, int size, SocketFlags socketFlags, out SocketError errorCode)
        {
            int num;
            if (this.CleanedUp)
            {
                throw new ObjectDisposedException(base.GetType().FullName);
            }
            if (buffer == null)
            {
                throw new ArgumentNullException("buffer");
            }
            if ((offset < 0) || (offset > buffer.Length))
            {
                throw new ArgumentOutOfRangeException("offset");
            }
            if ((size < 0) || (size > (buffer.Length - offset)))
            {
                throw new ArgumentOutOfRangeException("size");
            }
            errorCode = SocketError.Success;
            if (buffer.Length == 0)
            {
                num = OSSOCK.send(this.m_Handle.DangerousGetHandle(), null, 0, socketFlags);
            }
            else
            {
                fixed (byte* numRef = buffer)
                {
                    num = OSSOCK.send(this.m_Handle.DangerousGetHandle(), numRef + offset, size, socketFlags);
                }
            }
            if (num == -1)
            {
                errorCode = (SocketError)Marshal.GetLastWin32Error();
                this.UpdateStatusAfterSocketError(errorCode);
                return 0;
            }
            return num;
        }
        #endregion 
        #endregion

        #region sync shutdown and Close
        public void Shutdown(SocketShutdown how)
        {

            if (this.CleanedUp)
            {
                throw new ObjectDisposedException(base.GetType().FullName);
            }
            SocketError socketError = (OSSOCK.shutdown(this.m_Handle, (int)how) != SocketError.SocketError) ? SocketError.Success : ((SocketError)Marshal.GetLastWin32Error());
            switch (socketError)
            {
                case SocketError.Success:
                case SocketError.NotSocket:
                    this.SetToDisconnected();
                    this.InternalSetBlocking(this.willBlockInternal);
                    return;
            }
            SocketException socketException = new SocketException((int)socketError);
            this.UpdateStatusAfterSocketError(socketException);
            throw socketException;
        }
        public void Close()
        {
            ((IDisposable)this).Dispose();
        } 
        #endregion

        #region IDisposable Members

        void IDisposable.Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                int num;
                while ((num = Interlocked.CompareExchange(ref this.m_IntCleanedUp, 1, 0)) == 2)
                {
                    Thread.SpinWait(1);
                }
                if (num == 1)
                {
                    return;
                }
                this.SetToDisconnected();
                AsyncEventBits fdNone = AsyncEventBits.FdNone;
                if (this.m_BlockEventBits != AsyncEventBits.FdNone)
                {
                    //this.UnsetAsyncEventSelect();
                    //if (this.m_BlockEventBits == AsyncEventBits.FdConnect)
                    //{
                    //    LazyAsyncResult acceptQueueOrConnectResult = this.m_AcceptQueueOrConnectResult as LazyAsyncResult;
                    //    if ((acceptQueueOrConnectResult != null) && !acceptQueueOrConnectResult.InternalPeekCompleted)
                    //    {
                    //        fdNone = AsyncEventBits.FdConnect;
                    //    }
                    //}
                    //else if (this.m_BlockEventBits == AsyncEventBits.FdAccept)
                    //{
                    //    System.Collections.Queue queue = this.m_AcceptQueueOrConnectResult as Queue;
                    //    if ((queue != null) && (queue.Count != 0))
                    //    {
                    //        fdNone = AsyncEventBits.FdAccept;
                    //    }
                    //}
                }
                try
                {
                    int closeTimeout = this.m_CloseTimeout;
                    if (closeTimeout == 0)
                    {
                        this.m_Handle.Dispose();
                    }
                    else
                    {
                        SocketError error;
                        if (!this.willBlock || !this.willBlockInternal)
                        {
                            int argp = 0;
                            error = OSSOCK.ioctlsocket(this.m_Handle, -2147195266, ref argp);
                        }
                        if (closeTimeout < 0)
                        {
                            this.m_Handle.CloseAsIs();
                        }
                        else
                        {
                            error = OSSOCK.shutdown(this.m_Handle, 1);
                            unsafe
                            {
                                if (OSSOCK.setsockopt(this.m_Handle, SocketOptionLevel.Socket, SocketOptionName.ReceiveTimeout, ref closeTimeout, 4) != SocketError.Success)
                                {
                                    this.m_Handle.Dispose();
                                }
                                else if (OSSOCK.recv(this.m_Handle.DangerousGetHandle(), null, 0, SocketFlags.None) != 0)
                                {
                                    this.m_Handle.Dispose();
                                }
                                else
                                {
                                    int num4 = 0;
                                    if ((OSSOCK.ioctlsocket(this.m_Handle, 0x4004667f, ref num4) != SocketError.Success) || (num4 != 0))
                                    {
                                        this.m_Handle.Dispose();
                                    }
                                    else
                                    {
                                        this.m_Handle.CloseAsIs();
                                    }
                                } 
                            }
                        }
                    }
                }
                catch (ObjectDisposedException)
                {
                }
                //if (this.m_Caches != null)
                //{
                //    OverlappedCache.InterlockedFree(ref this.m_Caches.SendOverlappedCache);
                //    OverlappedCache.InterlockedFree(ref this.m_Caches.ReceiveOverlappedCache);
                //}
                //switch (fdNone)
                //{
                //    case AsyncEventBits.FdConnect:
                //        ThreadPool.UnsafeQueueUserWorkItem(new WaitCallback(((LazyAsyncResult)this.m_AcceptQueueOrConnectResult).InvokeCallback), new SocketException(SocketError.OperationAborted));
                //        goto Label_0253;

                //    case AsyncEventBits.FdAccept:
                //        ThreadPool.UnsafeQueueUserWorkItem(new WaitCallback(this.CompleteAcceptResults), null);
                //        goto Label_0253;
                //}
            }
            else
            {
                if ((this.m_Handle != null) && !this.m_Handle.IsInvalid)
                {
                    this.m_Handle.Dispose();
                }
                return;
            }
        Label_0253:
            if (this.m_AsyncEvent != null)
            {
                this.m_AsyncEvent.Close();
            }
        }
        private AsyncEventBits m_BlockEventBits;
        private ManualResetEvent m_AsyncEvent;
 


        #endregion

        #region Uitils Methods
        /// <summary>
        /// 获取SocketAddress的两个私有成员
        /// </summary>
        /// <param name="socketAddress"></param>
        /// <param name="m_Size"></param>
        /// <returns></returns>
        private byte[] GetM_buffer(SocketAddress socketAddress, out int m_Size)
        {
            byte[] m_buffer;
            FieldInfo socketAddress_m_Buffer = null;
            FieldInfo socketAddress_m_Size = null;
            socketAddress_m_Buffer = typeof(SocketAddress).GetField("m_Buffer",
                                                                     BindingFlags.Instance | BindingFlags.NonPublic);
            socketAddress_m_Size = typeof(SocketAddress).GetField("m_Size",
                                                                   BindingFlags.Instance | BindingFlags.NonPublic);
            m_buffer = (byte[])socketAddress_m_Buffer.GetValue(socketAddress);
            m_Size = (int)socketAddress_m_Size.GetValue(socketAddress);
            return m_buffer;
        }

        internal void InternalSetBlocking(bool desired)
        {
            bool flag;
            this.InternalSetBlocking(desired, out flag);
        }

        private SocketError InternalSetBlocking(bool desired, out bool current)
        {
            SocketError notSocket;

            int argp = desired ? 0 : -1;
            try
            {
                notSocket = OSSOCK.ioctlsocket(this.m_Handle, -2147195266, ref argp);
                if (notSocket == SocketError.SocketError)
                {
                    notSocket = (SocketError)Marshal.GetLastWin32Error();
                }
            }
            catch (ObjectDisposedException)
            {
                notSocket = SocketError.NotSocket;
            }
            if (notSocket == SocketError.Success)
            {
                this.willBlockInternal = argp == 0;
            }
            current = this.willBlockInternal;
            return notSocket;
        }

        internal void SetToConnected()
        {
            if (!this.m_IsConnected)
            {
                this.m_IsConnected = true;
                this.m_IsDisconnected = false;
            }
        }
        internal void UpdateStatusAfterSocketError(SocketError errorCode)
        {
            if (this.m_IsConnected && (this.m_Handle.IsInvalid || (((errorCode != SocketError.WouldBlock) && (errorCode != SocketError.IOPending)) && (errorCode != SocketError.NoBufferSpaceAvailable))))
            {
                this.SetToDisconnected();
            }
        }
        internal void UpdateStatusAfterSocketError(SocketException socketException)
        {
            this.UpdateStatusAfterSocketError((SocketError)socketException.NativeErrorCode);
        }
        internal void SetToDisconnected()
        {
            if (this.m_IsConnected)
            {
                this.m_IsConnected = false;
                this.m_IsDisconnected = true;
                if (!this.CleanedUp)
                {
                    //this.UnsetAsyncEventSelect();
                }
            }
        }
        internal bool CleanedUp
        {
            get
            {
                return (this.m_IntCleanedUp == 1);
            }
        }
        private void ValidateBlockingMode()
        {
            if (this.willBlock && !this.willBlockInternal)
            {
                throw new InvalidOperationException("net_invasync");
            }
        }
        #endregion
    }
}